require 'defs'
require 'utils'

job = ""
nan = 0./0.
-- Import all math functions
for k, v in pairs(math) do
	_G[k] = v
end
for k, v in pairs(bit) do
	_G[k] = v
end

-- Import all constants
for k, v in pairs(_sysc) do
	_G[k] = v
end

-- String -> numerical value conversion for constants
elements = {period=_sysc.PER, 
	mass=_sysc.MASS, ma=_sysc.MA,
	ecc=_sysc.ECC, lop=_sysc.LOP,
	inc=_sysc.INC, node=_sysc.NODE, radius=_sysc.RADIUS,
	ecc_anomaly = _sysc.ECCANOMALY, true_anomaly = _sysc.TRUEANOMALY}

local intmethods = {["Keplerian"] = _sysc.KEPLER,
	["Runge-Kutta 5"] = _sysc.RK5, ["Runge-Kutta 8"] = _sysc.RK8, ["Bulirsch-Stoer"] = _sysc.SWIFT_BS,
	["Wisdom-Holman"] = _sysc.SWIFT_WHM, ["RMVS"] = _sysc.SWIFT_RMVS3}

function _noticef(str, ...)
	io.stderr:write(string.format("!! " .. str, ...))
	io.stderr:write("\n")
	io.stderr:write(debug.traceback())
	io.stderr:flush()
end

_nassert = assert

vector = {}

matrix = {
	__class = "matrix",
	__index = function(m, k)
		if type(k) == "number" then
			error("Use the syntax m(i, j) to get the element [i,j]")
		else
			return matrix[k]
		end
	end,
	__newindex = function(m, k, v)
		error("Use the syntax m(i, j, val) to set the element [i,j] to val")
	end,
	__call = function(m, i, j, v)
		i = i - 1
		j = j - 1
		if v then
			m._h.data[i * m.__cols + j] = v
		else
			return m._h.data[i * m.__cols + j]
		end
	end,
	__tostring = function(m)
		local b = buf.buffer()
		for i = 1, m:rows() do
			for j = 1, m:cols() do
				b:appendf("%10.5e ", m(i, j))
			end
			b:append("\n")
		end
		return b:tostring()
	end,
	rows = function(m)
		return m.__rows
	end,
	cols = function(m)
		return m.__cols
	end,
	save = function(m, fn, type, fmt)
		local fid = ffi.C.fopen(fn, "w")
		_nassert(fid, "Could not open %s", fn)
		type = type or "text"
		fmt = fmt or "%e "
		if type == "binary" then
			ffi.C.fwrite(m._h.data, ffi.sizeof("double"), m:rows() * m:cols(), fid)
		else
			_sys.ok_fprintf_matrix(m._h, fid, fmt)
		end
		ffi.C.fclose(fid)
	end,
	new = function(r, c)
		if type(r) == 'number' then
			_nassert(type(c) == 'number', "Number of columns should be a number")
			local m = {_h = ffi.gc(_sys.gsl_matrix_alloc(r, c), _sys.gsl_matrix_free),
				__rows = r, __cols = c}
			setmetatable(m, matrix)
			return m
		elseif type(r) == "cdata" then
			local _h
			if c then
				_h = ffi.gc(r, _sys.gsl_matrix_free)
			else
				_h = r
			end
			local m = {_h = r, __rows = tonumber(r.size1), __cols = tonumber(r.size2)}
			setmetatable(m, matrix)
			return m
		end
	end
}

local bor_t = function(t, from)
	if type(t) == "number" then return t end
	local r = 0
	for i = 1, #t do
		if type(t[i]) == "string" then
			t[i] = from[t[i]]
		end
		r = bit.bor(r, t[i])
	end
	return r
end

kernel = {
	__class = "kernel", 
	__elements_mt = {
		__newindex = function(t, np, v)
			_nassert(np > 0 and np <= t._k:nplanets(), "Index outside bounds")		
			t._k:__set(np, t._el, v)
		end,
		__index = function(t, np)
			_nassert(np > 0 and np <= t._k:nplanets(), "Index outside bounds")		
			return _sys.K_getElement(t._k._h, np, t._el)
		end,
	},
	__offs_mt = {
		__newindex = function(t, nd, v)
			t._k:__set_data(nd, v)
		end,
		__index = function(t, nd)
			return _sys.K_getPar(t._k._h, nd-1)
		end
	},
	__index = function(k, a)
		if elements[a] then
			if rawget(k, '_' .. a) then
				return k['_' .. a]
			else
				k['_' .. a] = setmetatable({_k = k, _el = elements[a]}, kernel.__elements_mt)
				return k['_' .. a]
			end
		end
	
		return kernel[a]
	end,
	__tostring = function(k)
		local b = buf.buffer()
		b:append("Planets: ")
		b:append(tostring(matrix.new(_sys.K_getElements(k._h))))
		b:appendf("\nepoch = %.5f\n", k:epoch())
		b:appendf("chi2 = %e\n", k:chi2())
		b:appendf("rms = %e\n", k:rms())
		b:appendf("jitter = %e\n", k:jitter())
		b:appendf("int_method = %d\n", k:int_method())		
		b:appendf("ndata = %d\n", k:ndata())		
		b:appendf("nrvs = %d\n", k:nrvs())	
		b:appendf("ntts = %d\n", k:ntts())								
		return b:tostring()
	end,
	__set = function(k, pl, el, v)
		_nassert(pl >= 1, "Index should be bigger than 1")
		_nassert(pl <= k:nplanets(), "Index should be smaller than the number of planets loaded")
		_sys.K_setElement(k._h, pl, el, v)
	end,
	__set_data = function(k, d, v)
		_nassert(d >= 1, "Index should be bigger than 1")
		_nassert(d <= k:ndata(), "Index should be smaller than the number of datasets loaded")
		_sys.K_setPar(k._h, d-1, v)
	end,
	new = function(props, skip)
		local k
		if not props then
			k = {_h = ffi.gc(_sys.K_alloc(), _sys.K_free)}
		elseif type(props) == "table" and props.__class == "kernel" then
			k = {_h = ffi.gc(_sys.K_clone(props._h), _sys.K_free)}
		elseif type(props) == "string" then
			skip = skip or 1
			local fid = ffi.C.fopen(props, "r")
			_nassert(fid, "Could not open %s", props)
			local hload = _sys.K_load(fid, skip-1)
			if hload == nil then
				return nil
			end
			local _h = ffi.gc(hload,
				_sys.K_free)
			ffi.C.fclose(fid)			
			if not _h then
				return nil
			end
			k = {_h = _h}
		end
		setmetatable(k, kernel)
		k.offs = setmetatable({_k = k}, kernel.__offs_mt)		
		return k
	end,
	clone = function(k)
		return kernel.new(k)
	end,
	trange = function(k)
		if k._h.ndata <= 0 then
			return
		end
		local from = ffi.new("double[1]", 0)
		local to = ffi.new("double[1]", 0)
		_sys.K_getRange(k._h, from, to)
		return from[0], to[0]
	end,
	calculate = function(k)
		_sys.K_calculate(k._h)
	end,
	add = function(k, els) 
		local p = {}
		for k, v in pairs(els) do
			if elements[k] then
				p[#p+1] = elements[k]
				p[#p+1] = v
			end
		end
		p[#p+1] = _sysc.DONE

		_sys.K_addPlanet(k._h, ffi.new("const double[?]", #p, p))
		_sys.K_setElementRange(k._h, 1, ECC, -1, 0.95)
		--_sys.K_setElementRange(k._h, -1, MASS, 1e-4, 0./0.)		
	end,
	remove = function(k, idx)
		_nassert(idx >= 1, "Index should be bigger than 1")
		_nassert(idx <= k:nplanets(), "Index should be smaller than %d", k:nplanets())
		_sys.K_removePlanet(k._h, idx)
		_sys.K_compileData(k._h)
	end,
	_add_data = function(k, fn, type)
		if not ex.exists(fn) then
			_noticef("Could not open file %s", fn)
		elseif type then
			_sys.K_addDataFile(k._h, fn, type)
		elseif fn:find("%.vels") then
			_sys.K_addDataFile(k._h, fn, _sysc.T_RV)
		elseif fn:find("%.tds") or fn:find("%.tt") then
			_sys.K_addDataFile(k._h, fn, _sysc.T_TIMING)
		else 
			_noticef("Unknown data type for %s", fn)
		end
		k:data()
		return k._h.nsets
	end,
	add_data = function(k, fn, datatype)
		if type(fn) == "string" then
			return k:_add_data(fn, datatype)
		elseif type(fn) == "table" then
			for i = 1, #fn do 
				k:_add_data(fn[i], datatype) 
			end
			return k._h.nsets
		end
	end,
	remove_data = function(k, idx) 
		idx = idx or -1
		if idx == -1 then
			_sys.K_removeData(k._h, -1)
		elseif idx <= k._h.nsets then
			_sys.K_removeData(k._h, idx - 1)
			return true
		else
			_noticef("Index out of range (%d data sets loaded)", k._h.nsets)
		end
	end,
	int_method = function(k, intm)
		if intm then
			if type(intm) == "string" then
				_nassert(intmethods[intm], "Unknown integration method " .. intm)
				intm = intmethods[intm]
			end
			_sys.K_setIntMethod(k._h, intm)			
		end
		return _sys.K_getIntMethod(k._h)
	end,
	nplanets = function(k)
		return k._h.system.nplanets
	end,
	nsets = function(k)
		return k._h.nsets
	end,
	ndata = function(k, dtype)
		_sys.K_compileData(k._h)
		
		if not dtype then 
			return k._h.ndata
		else 
			local n = 0
			for i = 0, k._h.ndata-1 do
				if k._h.compiled[i][_sysc.T_FLAG] == dtype then
					n = n + 1
				end
			end
			return n
		end
	end,
	nrvs = function(k)
		return k._h.nrvs
	end,
	ntts = function(k)
		return k._h.ntts
	end,
	nrpars = function(k)
		return k._h.nrpars
	end,
	epoch = function(k, epoch)
		if epoch then
			_sys.K_setEpoch(k._h, epoch)
		end
		return k._h.system.epoch
	end,
	chi2 = function(k)
		return _sys.K_getChi2(k._h)
	end,
	rms = function(k)
		return _sys.K_getRms(k._h)
	end,
	jitter = function(k)
		return _sys.K_getJitter(k._h)
	end,
	chi2_rvs = function(k)
		return k._h.chi2_rvs
	end,
	chi2_tts = function(k)
		return k._h.chi2_tts
	end,
	rms_tts = function(k)
		return k._h.rms_tts
	end,
	mstar = function(k, mstar)
		if mstar then
			_sys.K_setMstar(k._h, mstar)
		end
		return _sys.K_getMstar(k._h)
	end,
	save = function(k, fn)
		local fid = ffi.C.fopen(fn, "w")
		if not fid then
			return nil
		end
		_sys.K_save(k._h, fid)
		ffi.C.fclose(fid)
		return true
	end,
	periodogram = function(k, per_type, samples, pmin, pmax, per_data)
		_nassert(k:nrvs() > 0, "No RV data loaded")
		per_data = per_data or _sysc.T_RV
		per_type = per_type or "all"
		samples = samples or 20000
		pmin = pmin or 0.5
		pmax = pmax or 2e4
		
		local valcol = _sysc.T_SVAL
		if per_type == "res" then
			valcol = _sysc.T_PRED
		end
		
		local m = _sys.K_combineData(k._h, per_data)
		local per = _sys.ok_periodogram_ls(m, samples, pmin, pmax, 0, _sysc.T_TIME,
			valcol, _sysc.T_ERR, nil)
		
		_sys.gsl_matrix_free(m)
		return matrix.new(per)
	end,
	flag = function(k, row, column, v)
		if row == "offs" then
			if v then
				_sys.K_setParFlag(k._h, column-1, v)
			end
			return _sys.K_getParFlag(k._h, column-1)
		else
			column = (type(column) == "number") and column or elements[column]
			if v then
				_sys.K_setElementFlag(k._h, row, column, v)
			end
			return _sys.K_getElementFlag(k._h, row, column)
		end
	end,
	minimize = function(k, opts)
		job = "Minimization"
		_nassert(k:ndata() > 0, "No data loaded")
		
		if opts then	
			local iters = opts.iters or 5000
			_sys.K_minimize(k._h, SIMPLEX, iters, nil)
		else
			k:calculate()		
			local chi2 = k:chi2()
			local i = 1
			while true do
				job = "Minimization pass " .. i
				local s = _sys.K_minimize(k._h, SIMPLEX, 5000, nil)
				if s ~= 0 then break end
				if k:chi2() - chi2 < -0.1 then
					chi2 = k:chi2()
				else 
					break
				end
				i = i + 1
			end
		end
		
		k:calculate()
	end,
	minimize1d = function(k, row, column, opts)
		job = "1-D minimization"
		_nassert(k:ndata() > 0, "No data loaded")	
		opts = opts or {iters = 10000}
		local iters = opts.iters or 10000
		if row == "offs" then
			row = -1
		end
		column = type(column) == "number" and column or elements[column]

		_sys.K_1dminimize(k._h, SIMPLEX, iters, row, column, nil)
		k:calculate()
	end, 
	data = function(k, nset)
		_nassert(k:ndata() > 0, "No data loaded")
		if not nset then
			return matrix.new(_sys.ok_buf_to_matrix(_sys.K_compileData(k._h), k:ndata(), _sysc.DATA_SIZE))
		else
			
		end
	end,
	range = function(k, row, column, minmax) 
		
		if row == "offs" then
			if minmax then
				_sys.K_setParRange(k._h, column-1, minmax.min or nan, minmax.max or nan)
			end
			
			local from = ffi.new("double[1]", 0)
			local to = ffi.new("double[1]", 0)
			_sys.K_getParRange(k._h, column - 1, from, to)
			return from[0], to[0]
		else
			_nassert(row > 0 and row <= t._k:nplanets(), "Index outside bounds")		
			
			column = (type(column) == "number") and column or elements[column]
			if minmax then
				_sys.K_setElementRange(k._h, row, column, minmax.min or nan, minmax.max or nan)
			end
			local from = ffi.new("double[1]", 0)
			local to = ffi.new("double[1]", 0)
			_sys.K_getElementRange(k._h, row, column, from, to)
			return from[0], to[0]
		end
	end
	
}


function acknowledgements() 
	local ack = [==[
TODO
]==]
	print(ack)
end